#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <ctype.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <getopt.h>
#include <string.h>
#include <errno.h>

#include "libcli.h"

/*#define LIBCLI_DEBUG*/

#ifdef LIBCLI_DEBUG
#define DBG(format, ...)  printf("DBG: %s: "format"\n", __FUNCTION__, ##__VA_ARGS__)
#else
#define DBG(format, ...)
#endif


int cli_question_yesno(const char *q, int defanswer)
{
  char answer[4];

  fprintf(stdout,"%s [%s]: ", q, defanswer?"Y/n":"y/N");
  fgets(answer, 4, stdin);

  if ((strlen(answer)==0))
    return defanswer;

  if ((answer[0]=='y') || (answer[0] == 'Y'))
    return 1;

  return 0;
}

void cli_mac2s(uint8_t * mac, char *s)
{
  int i;
  for (i = 0; i < 5; ++i)
    sprintf(&s[3*i], "%02X:", (uint16_t)mac[i]);
  sprintf(&s[15], "%02X", (uint16_t)mac[5]);
}

int cli_check_hex_digit(char *s)
{
  while (*s) {
    if (!isxdigit(*s))
      return 1;
    s++;
  }
  return 0;
}

int cli_s2n16(char *s, uint16_t *n)
{
  char *p;
  unsigned long int r;

  if (cli_check_hex_digit(s))
    return 1;  

  errno=0;
  r = strtoul(s, &p, 16);

  if (errno)
    return 1;

  if (*p)
    return 1;

  if (r>65535)
    return 1;

  (*n)=r;

  return 0;
}

int cli_s2nl16(char *s, uint32_t *n)
{
  char *p;

  if (cli_check_hex_digit(s))
    return 1;  

  errno=0;
  (*n) = strtoul(s, &p, 16);

  if (errno)
    return 1;

  if (*p)
    return 1;

  return 0;
}

char *cli_strchr_nothex(char *s)
{
  int i=0;
  while (s[i]) {
    if (!isxdigit(s[i]))
      return &s[i];
    i++;
  }
  return 0;
}

int cli_s2mac(uint8_t * mac, char *s)
{
  int len, sidx, eidx, i;
  char *c, ch=0;
  uint16_t n;

  len = strlen(s);

  if (len < 11)
    return 1;

  sidx = 0;

  for (i = 0; i < 6; ++i) {

    /* find delimeter */

    if (i < 5) {
      c = cli_strchr_nothex(&s[sidx]);
      /*c = strchr(&s[sidx], ':');*/
      if (c)
        eidx = (int)(c - s);
      else
        return 1;
    } else
       eidx = (len-1);

   /* printf("idx: %i\n",sidx); */

   /* read hex number */

   if (i < 5) {
     ch = s[eidx];
     s[eidx]=0;
   }

   if (cli_s2n16(&s[sidx], &n))
     return 1;

   if (i<5)
    s[eidx]=ch;

   if (n>255)
     return 1;

   mac[i] = (uint8_t)n;

   /* next */

   if ((eidx+1) < len)
     sidx = (eidx+1);

 }


 return 0;

}


int cli_check_bin_digit(char *s)
{
  if (!s)
     return 1;
  while (*s) {
    if (((*s)!='0') && ((*s)!='1'))
      return 1;
    s++;
  }
  return 0;
}

int cli_s2n2(char *s, uint16_t *n)
{
  char *p;
  unsigned long int r;
  if (cli_check_bin_digit(s))
    return 1;

  errno=0;
  r = strtoul(s, &p, 2);

  if (errno)
    return 1;

  if (*p)
    return 1;

  if (r>65535)
    return 1;

  (*n)=r;

  return 0;
}

int cli_s2nl2(char *s, uint32_t *n)
{
  char *p;
  if (cli_check_bin_digit(s))
    return 1;

  errno=0;
  (*n) = strtoul(s, &p, 2);

  if (errno)
    return 1;

  if (*p)
    return 1;

  return 0;
}

int cli_check_dec_digit(char *s)
{
  if (!s)
     return 1;
  while (*s) {
    if (!isdigit(*s))
      return 1;
    s++;
  }
  return 0;
}

int cli_s2n10(char *s, uint16_t *n)
{
  char *p;
  unsigned long int r;
  if (cli_check_dec_digit(s))
    return 1;

  errno=0;
  r = strtoul(s, &p, 10);

  if (errno)
    return 1;

  if (*p)
    return 1;

  if (r>65535)
    return 1;

  (*n)=r;

  return 0;
}

int cli_s2nl10(char *s, uint32_t *n)
{
  char *p;
  if (cli_check_dec_digit(s))
    return 1;

  errno=0;
  (*n) = strtoul(s, &p, 10);

  if (errno)
    return 1;

  if (*p)
    return 1;

  return 0;
}

void cli_changecase(char *s, char up)
{
  while (*s) {
    (*s)=(up)?toupper(*s):tolower(*s);
    s++;
  }
}

int cli_hexdump(FILE *f, void *buf, size_t size, uint32_t addr)
{
  uint32_t i, j;

  for(i = 0; i < size; i += 16) {

    fprintf(f, "%04X : ", i+addr);

    for (j = 0; j < 16 && i + j < size; j++)
      fprintf(f, "%2.2X ", ((uint8_t*)buf)[i + j]);

    for (; j < 16; j++)
      fprintf(f, "   ");

    fprintf(f,": ");
    for (j = 0; j < 16 && i + j < size; j++) {
      char c = toascii(((uint8_t*)buf)[i + j]);
      fprintf(f, "%c", isalnum(c) ? c : '.');
    }

    fprintf(f, "\n");
  }

  return 0;
}

int cli_ip2s(uint8_t *d, char *s)
{
  sprintf(s, "%d.%d.%d.%d", (uint16_t)d[0], (uint16_t)d[1], (uint16_t)d[2], (uint16_t)d[3]);
  return 0;
}


int cli_s2ip(uint32_t *a, char *s)
{
  struct in_addr addr;

  if ( (!s) || (!s[0]) )
    return -1;

  if (!inet_aton(s, &addr))
     return -1;

  (*a) = addr.s_addr;
  return 0;
}

int cli_s2mask(uint32_t *m, char *s)
{
  struct in_addr addr;

  if ( (!s) || (!s[0]) )
    return -1;

  if (!inet_aton(s, &addr))
    return -1;
  (*m) = addr.s_addr;
  return 0;
}

int cli_add_options2loptions2(struct option *opt, cli_option_item_t **o)
{
  uint16_t i,j;

  i=0;
  while (opt[i].name) i++;

  j=0;
  while (o[j]) {
    DBG("lopt:%u opt:%u n:'%s'\n",i, j, o[i]->name);
    opt[i].name = o[i]->name;
    opt[i].has_arg = (o[i]->type!=CLI_OPTION_FLAG)?1:0;
    opt[i].flag = 0;
    opt[i].val = 0;
    i++;j++;
  }

  memset(&opt[i], 0, sizeof(struct option));

  return 0;
}



#define MAXOPTIONS 64

struct option *cli_create_loption_array(uint16_t max)
{
  struct option *opt=0;

  if (!max)
    max = MAXOPTIONS;

  DBG("max %u", max);

  opt = malloc(sizeof(struct option)*(max+1));
  if (!opt)
    return 0;

  memset(opt, 0, sizeof(struct option)*(max+1));

  return opt;
}

void cli_free_loption_array(struct option *o)
{
  if (o)
    free(o);
}

cli_option_item_t* cli_create_option_array(size_t max)
{
  cli_option_item_t *opt;

  opt = malloc(sizeof(cli_option_item_t)*((max)?max:MAXOPTIONS)+1);
  if (!opt)
    return 0;

  memset(opt, 0, sizeof(cli_option_item_t));

  return opt;
}

void cli_free_option_array(cli_option_item_t *o)
{
  if (o) {
    free(o);
    o=0;
  }
}

cli_option_item_t** cli_create_option_array2(size_t max)
{
  cli_option_item_t **opt;

  opt = malloc(sizeof(cli_option_item_t*)*((max)?max:MAXOPTIONS)+1);
  if (!opt)
    return 0;

  memset(opt, 0, sizeof(cli_option_item_t*)*((max)?max:MAXOPTIONS)+1);

  return opt;
}

void cli_free_option_array2(cli_option_item_t** o)
{
  if (o)
    free(o);
}


uint16_t cli_opt_size2(cli_option_item_t **o)
{
  uint16_t i=0;
  while (o[i]) i++;
  return i;
}

int cli_add_options2(cli_option_item_t **to, const cli_option_item_t *from)
{
  uint16_t i,j;

  i = cli_opt_size2(to);

  if (!from)
    return 0;

  j=0;
  while (from[j].name) {
    to[i] = (cli_option_item_t *)&from[j];
    /*memcpy(&to[i], &from[j], sizeof(cli_option_item_t));*/
    i++;j++;
  }

  to[i]=NULL;

  return 0;
}

uint16_t cli_opt_size(cli_option_item_t *o)
{
  uint16_t i=0;
  while (o[i].name) i++;
  return i;
}

int cli_add_options(cli_option_item_t *to, const cli_option_item_t *from)
{
  uint16_t i,j;

  i = cli_opt_size(to);

  j=0;
  while (from[j].name) {
    memcpy(&to[i], &from[j], sizeof(cli_option_item_t));
    i++;j++;
  }

  memset(&to[i], 0, sizeof(struct option));

  return 0;
}

int cli_handle_long_opt(cli_option_item_t *opts, int idx, char *arg)
{
  int r = 0;
  /*printf("%s %s %04x\n",__FUNCTION__, opts[idx].name, opts[idx].flags);*/
  opts[idx].flags &= ~CLI_OPTIONFL_ACCEPTED;
  switch (opts[idx].type) {
    case CLI_OPTION_FLAG:
      *((uint8_t*)opts[idx].data) = 1;
      break;
    case CLI_OPTION_STR:
      strncpy((char *)opts[idx].data, arg, opts[idx].datasize);
      if (opts[idx].datasize)
        ((char *)opts[idx].data)[opts[idx].datasize-1] = 0;
      break;
    case CLI_OPTION_UINT16:
      if (cli_s2n10(arg, ((uint16_t*)opts[idx].data)))
        r=1;
      break;
    case CLI_OPTION_UINT32:
      if (cli_s2nl10(arg, ((uint32_t*)opts[idx].data)))
        r=1;
      break;
    case CLI_OPTION_UINT16X:
      if (cli_s2n16(arg, ((uint16_t*)opts[idx].data)))
        r=1;
      break;
    case CLI_OPTION_UINT32X:
      if (cli_s2nl16(arg, ((uint32_t*)opts[idx].data)))
        r=1;
      break;
    default:
      r=1;
      break;
  }

  if (r)
    printf("Wrong argument for '%s' option!\n", opts[idx].name);
  else
    opts[idx].flags |= CLI_OPTIONFL_ACCEPTED;

  /*printf("%s %s %04x\n",__FUNCTION__, opts[idx].name, opts[idx].flags);*/

  return r;
}

int cli_handle_long_opt2(cli_option_item_t **opts, int idx, char *arg)
{
  int r = 0;
  /*printf("%s %s %04x\n",__FUNCTION__, opts[idx].name, opts[idx].flags);*/
  opts[idx]->flags &= ~CLI_OPTIONFL_ACCEPTED;
  switch (opts[idx]->type) {
    case CLI_OPTION_FLAG:
      *((uint8_t*)opts[idx]->data) = 1;
      break;
    case CLI_OPTION_STR:
      strncpy((char *)opts[idx]->data, arg, opts[idx]->datasize);
      if (opts[idx]->datasize)
        ((char *)opts[idx]->data)[opts[idx]->datasize-1] = 0;
      break;
    case CLI_OPTION_UINT16:
      if (cli_s2n10(arg, ((uint16_t*)opts[idx]->data)))
        r=1;
      break;
    case CLI_OPTION_UINT32:
      if (cli_s2nl10(arg, ((uint32_t*)opts[idx]->data)))
        r=1;
      break;
    case CLI_OPTION_UINT16X:
      if (cli_s2n16(arg, ((uint16_t*)opts[idx]->data)))
        r=1;
      break;
    case CLI_OPTION_UINT32X:
      if (cli_s2nl16(arg, ((uint32_t*)opts[idx]->data)))
        r=1;
      break;
    default:
      r=1;
      break;
  }

  if (r)
    printf("Wrong argument for '%s' option!\n", opts[idx]->name);
  else
    opts[idx]->flags |= CLI_OPTIONFL_ACCEPTED;

  /*printf("%s %s %04x\n",__FUNCTION__, opts[idx].name, opts[idx].flags);*/

  return r;
}

int cli_print_options_help2(const cli_option_item_t **o)
{
  char s[33];
  while (*o) {

    switch ((*o)->type) {
      default:
      case CLI_OPTION_FLAG:
        snprintf(s, sizeof(s), "%s", (*o)->name);
      break;
      case CLI_OPTION_STR:
        snprintf(s, sizeof(s), "%s <string>", (*o)->name);
      break;
      case CLI_OPTION_UINT16: case CLI_OPTION_UINT32:
        snprintf(s, sizeof(s), "%s <num>", (*o)->name);
      break;
      case CLI_OPTION_UINT16X: case CLI_OPTION_UINT32X:
        snprintf(s, sizeof(s), "%s <hexnum>", (*o)->name);
      break;
    }
    if (!((*o)->flags&CLI_OPTIONFL_HIDDEN))
      printf("-%-24s - %s\n", s, (*o)->help);
    o++;
  }
  return 0;
}

int cli_print_args_help2(const cli_option_item_t **o)
{
  while (*o) {
    printf("[ %s ]  ", (*o)->help);
    o++;
  }
  printf("\n");
  return 0;
}

int cli_check_opt_req(cli_option_item_t *o)
{
  while (o->name) {
    if (o->flags & CLI_OPTIONFL_REQUIRED)
      if (!(o->flags & CLI_OPTIONFL_ACCEPTED))
        return 1;
    o++;
  }
  return 0;
}

int cli_check_opt_req2(const cli_option_item_t **o, cli_option_item_t **failo)
{
  DBG("");
  while (*o) {
    if ((*o)->flags & CLI_OPTIONFL_REQUIRED)
      if (!((*o)->flags & CLI_OPTIONFL_ACCEPTED)) {
        if (failo)
          (*failo) = (cli_option_item_t *)(*o);
        return 1;
      }
    o++;
  }
  return 0;
}


int cli_handle_opt2(cli_option_item_t **opts, cli_option_item_t **args, int argc, char **argv)
{
  int r=0, optidx, c;
  struct option * lopt;
  cli_option_item_t *o;

  DBG("");

  lopt = cli_create_loption_array(cli_opt_size2(opts));

  if (!lopt) {
    r = 1;
    goto exit;
  }

  cli_add_options2loptions2(lopt, opts);

  while (1) {

    c = getopt_long_only(argc, argv, "", lopt,  &optidx);

    DBG("c: %i optidx: %i", c, optidx);

    if (c==-1)
      break;

    switch (c) {
      case 0:
        if (cli_handle_long_opt2(opts, optidx, optarg)) {
          r = 1;
          goto exit;
        }
      break;
      case '?': case ':':
        r=1;
        goto exit;

      default: 
        r=1;
        goto exit;
    }
  }

  if (cli_check_opt_req2((const cli_option_item_t **)opts, &o)) {
    printf("Required option '-%s' missing !\n", o->name);
    r=1;
    goto exit;
  }

  if ((args) && (args[0])) {

    c = 0;
    while (args[c]) {
      if (((optind+c) < argc))
        if (cli_handle_long_opt2(args, c, argv[optind+c])) {
         r=1;
         goto exit;
      }
      c++;
    }

    if (cli_check_opt_req2((const cli_option_item_t **)opts, &o)) {
      printf("Required argument '%s' missing !\n", o->help);
      r=1;
      goto exit;
    }

    if ((optind+c) < argc)
      printf("Too much number of arguments!\n");

  }

exit:

  cli_free_loption_array(lopt);
  return r;
}


int _cli_is_opt_accepted(cli_option_item_t *o, void *pdata)
{
  while (o->name) {    
    if (o->data == pdata)
      return (o->flags & CLI_OPTIONFL_ACCEPTED);
    o++;
  }
  return 0;
}

int _cli_is_opt_accepted2(const cli_option_item_t **o, void *pdata)
{
  while (*o) {
    if ((*o)->data == pdata)
      return ((*o)->flags & CLI_OPTIONFL_ACCEPTED);
    o++;
  }
  return 0;
}

int cli_is_opt_accepted(cli_option_item_t *opt, cli_option_item_t *arg, void *pdata)
{
  int r = 0;

  if (opt)
    r |= _cli_is_opt_accepted(opt, pdata);
  if (arg)
    r |= _cli_is_opt_accepted(arg, pdata);
  return r;
}

int cli_is_opt_accepted2(const cli_option_item_t **opt, const cli_option_item_t **arg, void *pdata)
{
  int r = 0;

  if (opt)
    r |= _cli_is_opt_accepted2(opt, pdata);
  if (arg)
    r |= _cli_is_opt_accepted2(arg, pdata);
  return r;
}
